"""Built-in reference parsers."""

import contextlib
import os
import importlib
import pip
import shutil
import tempfile
import uuid
from zipfile import ZipFile

import boto3facade
from boto3facade.s3 import S3
from boto3facade.cloudformation import Cloudformation
import jinja2
from s3keyring.s3 import S3Keyring

from humilis.exceptions import ReferenceError, InvalidLambdaDependencyError
import humilis.utils as utils


def _get_s3path(layer, config, full_path):
    """Returns the S3 target (bucket, key) for a local file."""
    env_prefix = "{base_prefix}{env_name}/".format(
        base_prefix=config.profile.get('s3prefix', ''),
        env_name=layer.env_name)
    if layer.env_stage is not None:
        env_prefix = "{}{}/".format(env_prefix, layer.env_stage)

    s3key = "{env_prefix}{layer_name}/{file_name}".format(
        env_prefix=env_prefix,
        layer_name=layer.name,
        file_name=os.path.basename(full_path))
    s3bucket = config.profile.get('bucket')
    return (s3bucket, s3key)


def secret(layer, config, service=None, key=None, group=None, kms_key_id=None):
    """Retrieves a secret stored in a S3 keyring.

    :param service: An alias of group, for backwards compatibility
    :param key: The key used to identify the secret within the server
    :param group: The name of the group of secrets
    :param kms_key_id: The ID of the KMS Key to encrypt the secret

    :returns: The plaintext or encrypted secret
    """
    if not group:
        group = service

    s3keyring_config = os.path.join(layer.env_basedir, ".s3keyring.ini")
    if os.path.isfile(s3keyring_config):
        kr = S3Keyring(config_file=s3keyring_config)
    else:
        kr = S3Keyring()

    secret = kr.get_password(group, key)
    if kms_key_id:
        return boto3.client('kms').encrypt(KeyId=kms_key_id, Plaintext=secret)
    return secret


def file(layer, config, path=None):
    """Uploads a local file to S3 and returns the corresponding S3 path.

    :param layer: The Layer object for the layer declaring the reference.
    :param config: An object holding humilis configuration options.
    :param path: Path to the file, relative to the location of meta.yaml.

    :returns: The S3 path where the file has been uploaded.
    """
    full_path = os.path.join(layer.basedir, path)
    s3bucket, s3key = _get_s3path(layer, config, full_path)
    s3 = S3(config)
    s3.cp(full_path, s3bucket, s3key)
    layer.logger.info("{} -> {}/{}".format(full_path, s3bucket, s3key))
    if layer.type == "sam":
        return os.path.join("s3://", s3bucket, s3key)
    else:
        return {'s3bucket': s3bucket, 's3key': s3key}


def lambda_ref(layer, config, path=None, dependencies=None, **params):
    """Prepares a lambda deployment package and uploads it to S3.

    :param layer: The Layer object for the layer declaring the reference.
    :param config: An object holding humilis configuration options.
    :param path: Path to the file, relative to the location of meta.yaml.
    :param dependencies: A list of Python dependencies.

    :returns: S3 path where the deployment package has been uploaded.
    """
    fpath = os.path.abspath(os.path.join(layer.basedir, path))
    logger = layer.logger
    if os.path.isdir(fpath):
        with _deploy_package(fpath, layer, logger, dependencies, params) as fpath:
            s3path = file(layer, config, fpath)
    else:
        path_no_ext, ext = os.path.splitext(fpath)
        if ext == '.zip':
            s3path = file(layer, config, fpath)
        else:
            with _simple_deploy_package(fpath, layer, logger, params) as fpath:
                s3path = file(layer, config, fpath)

    return s3path


def _install_dependencies(layer, path, dependencies):
    """Install Python dependencies under the given path."""
    for dep in dependencies:
        deppath = os.path.abspath(os.path.join(layer.env_basedir, dep))
        targetpath = os.path.join(path, os.path.basename(dep))
        if os.path.isfile(deppath):
            ext = os.path.splitext(dep)[1]
            if ext == ".py":
                # A self-contained Python module
                shutil.copyfile(deppath, targetpath)
            elif ext == ".txt":
                # A requirements file
                pip.main(['install', '-r', deppath, '-t', path])
            else:
                raise InvalidLambdaDependencyError(dep)
        elif os.path.isdir(deppath):
            if os.path.isfile(os.path.join(deppath, "setup.py")):
                # A local pip installable
                pip.main(['install', deppath, '-t', path])
            else:
                # A self-contained Python package
                shutil.copytree(deppath, targetpath)
        else:
            if dep.find("git+") >= 0:
                # A git repo
                pip.main(['install', '-e', dep, '-t', path])
            elif dep.find(":") < 0:
                # A Pypi package
                pip.main(['install', dep, '-t', path, '--upgrade'])
            else:
                # A private index package
                index = ":".join(dep.split(":")[:-1])
                pip.main(['install', '-i', index, dep, '-t', path,
                          '--upgrade'])


@contextlib.contextmanager
def _deploy_package(path, layer, logger, dependencies, params):
    """Creates a deployment package for multi-file lambda with deps."""
    with utils.move_aside(path) as tmppath:
        # removes __* and .* dirs
        _cleanup_dir(tmppath)
        # render Jinja2 templated files
        template_params = layer.loader_params
        template_params.update(params)
        _preprocess_dir(tmppath, template_params)
        setup_file = os.path.join(tmppath, 'setup.py')
        if os.path.isfile(setup_file):
            # Install all depedendencies in the same dir
            pip.main(['install', tmppath, '-t', tmppath])
        requirements_file = os.path.join(tmppath, 'requirements.txt')
        if os.path.isfile(requirements_file):
            pip.main(['install', '-r', requirements_file, '-t', tmppath])

        if dependencies:
            _install_dependencies(layer, tmppath, dependencies)

        suffix = str(uuid.uuid4())
        tmpdir = tempfile.mkdtemp()
        basename = os.path.basename(path)
        zipfile = os.path.join(tmpdir, "{}{}{}".format(basename, suffix,
                                                       '.zip'))
        with ZipFile(zipfile, 'w') as myzip:
            utils.zipdir(tmppath, myzip)
        yield zipfile
        shutil.rmtree(tmpdir)


def _cleanup_dir(path):
    """Removes __* and .* dirs."""
    to_remove = []
    for root, dirs, files in os.walk(path):
        for dirpath in dirs:
            if dirpath.startswith('__') or dirpath.startswith('.'):
                to_remove.append(os.path.join(root, dirpath))
    for dirpath in to_remove:
        shutil.rmtree(dirpath, ignore_errors=True)


@contextlib.contextmanager
def _simple_deploy_package(path, layer, logger, params):
    """Creates a deployment package for a one-file no-deps lambda."""
    logger.info("Creating deployment package for '{}'".format(path))
    with utils.move_aside(path) as tmppath:
        template_params = layer.loader_params
        template_params.update(params)
        _preprocess_file(tmppath, template_params)
        path_no_ext, ext = os.path.splitext(tmppath)
        basename = os.path.basename(path_no_ext)
        gc = _git_head()
        suffix = ('-' + gc, '')[gc is None]
        tmpdir = tempfile.mkdtemp()
        zipfile = os.path.join(tmpdir,
                               "{}{}{}".format(basename, suffix, '.zip'))
        with ZipFile(zipfile, 'w') as myzip:
            myzip.write(tmppath, arcname=basename + ext)
        yield zipfile
        shutil.rmtree(tmpdir)


def _is_jinja2_template(path):
    """Returns true if a file contains a jinja2 template."""
    _, ext = os.path.splitext(path)
    if ext in {'.pyc'}:
        return False

    result = False
    with open(path, 'r') as f:
        for line in f:
            if line.startswith('#') and line.find('preprocessor:jinja2'):
                result = True
                break
    return result


def _preprocess_file(path, params):
    """Render in place a jinja2 template."""
    if not _is_jinja2_template(path):
        return
    basedir, filename = os.path.split(path)
    env = jinja2.Environment(
        extensions=["jinja2.ext.with_"],
        loader=jinja2.FileSystemLoader(basedir))
    # Add custom functions and filters
    utils.update_jinja2_env(env)
    result = env.get_template(filename).render(params)
    with open(path, 'w') as f:
        f.write(result)


def _preprocess_dir(path, params):
    """Preprocesses all files in a directory using Jinja2."""
    for root, dirs, files in os.walk(path):
        for file in files:
            filepath = os.path.join(root, file)
            _preprocess_file(filepath, params)


def layer(layer, config, layer_name=None, resource_name=None,
          output_name=None):
    """Gets the physical ID of a resource in an already deployed layer.

    :param layer: The Layer object for the layer declaring the reference.
    :param config: An object holding humilis configuration options.
    :param layer_name: The name of the layer that contains the target resource.
    :param resource_name: The logical name of the target resource.
    :param output_name: The name of the layer output.

    :returns: The physical ID of the resource.
    """
    if not (resource_name or output_name) or (resource_name and output_name):
        raise ValueError(
            "Exactly one of these two parameters should be provider: either"
            "'resource_name' or 'output_name'")

    if resource_name:
        stack_name = utils.get_cf_name(layer.env_name, layer_name,
                                       stage=layer.env_stage)
        return _get_stack_resource(layer, config, stack_name, resource_name)
    else:
        return output(layer, config, layer_name=layer_name,
                      output_name=output_name)



def environment(layer, config, environment_name=None, stage=None,
                layer_name=None, resource_name=None, output_name=None):
    """Gets the physical ID of a resource in another environment.

    :param layer: The Layer object of the layer declaring the reference.
    :param config: An object holding humilis configuration options.
    :param layer_name: The name of the layer that contains the target resource.
    :param resource_name: The logical name of the target resource.
    :param output_name: The name of the layer output

    :returns: The physical ID of the resource.
    """

    if not (resource_name or output_name) or (resource_name and output_name):
        raise ValueError(
            "Exactly one of these two parameters should be provider: either"
            "'resource_name' or 'output_name'")
    if resource_name:
        stack_name = utils.get_cf_name(environment_name, layer_name, stage=stage)
        return _get_stack_resource(layer, config, stack_name, resource_name)
    else:
        return output(
            layer, config, layer_name=layer_name, stage=stage,
            environment_name=environment_name, output_name=output_name)



def _get_stack_resource(layer, config, stack_name, resource_name):
    """Gets the physical ID of a resource in a CF Stack.

    :param stack_name: The name of the CF stack.
    :param resource_name: The logical name of the CF resource.

    :returns: The physical ID of the resource.
    """
    cf = Cloudformation(config)
    resource = cf.get_stack_resource(stack_name, resource_name)

    if len(resource) < 1:
        all_stack_resources = [x.logical_resource_id for x
                               in cf.get_stack_resources(stack_name)]
        msg = "{} does not exist in stack {} (with resources {}).".format(
            resource_name, stack_name, all_stack_resources)
        raise ReferenceError(resource_name, msg, logger=layer.logger)

    return resource[0].physical_resource_id


def output(layer, config, layer_name=None, output_name=None,
           environment_name=None, stage=None):
    """Gets the value of an output produced by an already deployed layer.

    :param layer: The Layer object for the layer declaring the reference.
    :param config: An object holding humilis configuration options.
    :param layer_name: The logical name of the layer that produced the output.
    :param output_name: The logical name of the output parameter.
    """
    if not environment_name:
        environment_name = layer.env_name

    if not stage:
        stage = layer.env_stage

    stack_name = utils.get_cf_name(environment_name, layer_name, stage=stage)
    cf = Cloudformation(config)
    try:
        output = cf.get_stack_output(stack_name, output_name)
    except AttributeError:
        msg = "No output '{}' in CF stack '{}'".format(output_name, stack_name)
        ref = "output/{}/{}/{}/{}".format(environment_name, layer_name, stage,
                                          output_name)
        raise ReferenceError(ref, msg)
    if len(output) < 1:
        all_stack_outputs = [x['OutputKey'] for x
                             in cf.stack_outputs[stack_name]]
        msg = ("{} output does not exist for stack {} "
               "(with outputs {}).").format(output_name,
                                            stack_name,
                                            all_stack_outputs)
        ref = "output ({}/{})".format(layer_name, output_name)
        raise ReferenceError(ref, msg, logger=layer.logger)
    return output[0]


def boto3(layer, config, service=None, call=None, output_attribute=None,
          output_key=None):
    """Calls a boto3facade method.

    :param layer: The Layer object for the layer declaring the reference.
    :param config: An object holding humilis configuration options.
    :param service: The name of the AWS service.
    :param call: A dict with two keys: method, and parameters.
    :param output_attribute: Object attribute to return.
    :param output_key: Dictionary key to return.

    :returns: The call response, or its corresp. attribute or key.
    """
    facade_name = service.title()
    if not hasattr(boto3facade, service):
        ref = "boto3facade.{}.{}.{}: {}".format(service, facade_name,
                                                call['method'],
                                                call['parameters'])
        msg = "Service {} not supported".format(service)
        raise ReferenceError(ref, msg, logger=layer.logger)

    module = importlib.import_module("boto3facade.{}".format(service))
    facade_cls = getattr(module, facade_name)
    facade = facade_cls(config)
    method = getattr(facade, call['method'])
    args = call.get('args', [])
    kwargs = call.get('kwargs', {})
    result = method(*args, **kwargs)
    if not isinstance(result, str) and not isinstance(result, dict) and \
        hasattr(result, '__iter__'):
            # Convert iterables to lists
            result = list(result)
    if isinstance(result, list):
        if len(result) == 1:
            result = result[0]
        else:
            raise ReferenceError("boto3/{}/{}".format(service, call),
                "Must produce exactly one result but {} were produced".format(
                len(result)))

    if output_attribute is not None:
        result = getattr(result, output_attribute)
    if output_key is not None:
        return result.get(output_key)
    else:
        return result


def j2_template(layer, config, path=None, s3_upload=False, params=None):
    """Render a j2 template and return the local or s3 path of the result.

    :param layer: The layer object for the layer declaring the reference.
    :param config: An object holding humilis configuration options.
    :param path: The path of the j2 template to render. Relative to meta.yml.
    :param s3_upload: Upload the rendered template to s3 or not.
    :param params: A dict containing the values to render the template.

    :returns: The local or s3 path of the rendered template.
    """
    if params is None:
        msg = ("Missing params for j2 rendering in layer '{}' "
               "and env '{}'").format(layer.name, layer.env_name)
        ref = "j2_template '{}')".format(path)
        raise ReferenceError(ref, msg, logger=layer.logger)
    basefile, j2_ext = os.path.splitext(path)
    if j2_ext not in {".j2"}:
        msg = ("The file is not a Jinja2 template. layer : '{}', "
               "env : '{}'".format(layer.name, layer.env_name))
        ref = "j2_template '{}')".format(path)
        raise ReferenceError(ref, msg, logger=layer.logger)

    _, ext = os.path.splitext(basefile)
    _, filename = os.path.split(path)
    env = jinja2.Environment(
        extensions=["jinja2.ext.with_"],
        loader=jinja2.FileSystemLoader(layer.basedir))
    result = env.get_template(filename).render(params)
    output_path = os.path.join(layer.env_basedir, "ref-j2_template_" +
                               str(uuid.uuid4()) + ext)
    with open(output_path, "w") as f:
        f.write(result)
    if s3_upload:
        s3bucket, s3key = _get_s3path(layer, config, output_path)
        s3 = S3(config)
        s3.cp(output_path, s3bucket, s3key)
        layer.logger.info("{} -> {}/{}".format(output_path, s3bucket, s3key))
        return os.path.join("s3://", s3bucket, s3key)

    return output_path



"""Humilis Layer."""

import itertools
import os
import os.path
import re
import logging
import time
import uuid
from humilis.config import config
from humilis.utils import DirTreeBackedObject, get_cf_name
from humilis.exceptions import (ReferenceError, CloudformationError,
                                MissingPluginError)
import boto3
from boto3facade.s3 import S3
from boto3facade.ec2 import Ec2
from boto3facade.cloudformation import Cloudformation
from boto3facade.exceptions import NoUpdatesError
from botocore.exceptions import ClientError
import json
import time
import datetime
from uuid import uuid4


def _is_legacy_reference(value):
    """True if a parameter value is a reference using legacy syntax."""
    return isinstance(value, dict) and 'ref' in value and \
        'parser' in value["ref"]


def _is_reference(value):
    """True if a parameter value is a reference."""
    return isinstance(value, dict) and len(value) == 1 and \
        list(value.keys())[0][0] == '$'


class Layer:
    """A layer of infrastructure that translates into a single CF stack"""
    def __init__(self, __env, __name, layer_type=None, logger=None,
                 loader=None, humilis_profile=None, **user_params):
        self.__environment_repr = repr(__env)
        self.environment = __env
        if not humilis_profile:
            self.cf = self.environment.cf
        else:
            config.boto_config.activate_profile(humilis_profile)
            self.cf = Cloudformation(config.boto_config)
        if logger is None:
            self.logger = logging.getLogger(__name__)
            # To prevent warnings
            self.logger.addHandler(logging.NullHandler())
        else:
            self.logger = logger
        self.name = __name
        self.env_name = self.environment.name
        self.env_stage = self.environment.stage
        self.env_basedir = self.environment.basedir
        self.depends_on = []
        self.section = {}
        self.type = layer_type
        self.s3_prefix = "{base}{env}/{stage}/{layer}/".format(
            base=config.boto_config.profile.get("s3prefix"),
            env=self.environment.name,
            stage=self.environment.stage,
            layer=__name)

        if layer_type is not None:
            basedir = config.layers.get(layer_type)
            if not basedir:
                msg = ("The plugin providing the layer type '{}' is not "
                       "installed in this system. Please install it and "
                       "try again.").format(layer_type)
                raise MissingPluginError(msg)
        else:
            basedir = None

        if basedir is None:
            basedir = os.path.join(self.env_basedir, 'layers', self.name)

        self.basedir = basedir

        if loader is None:
            loader = DirTreeBackedObject(basedir, self.logger)

        self.loader = loader

        # These param set will be sent to the template compiler and will be
        # populated once the layers this layer depend on have been created.
        self.params = {}

        # the parameters that will be used to compile meta.yaml
        self.meta = {}
        meta_params = {p[0]: p[1] for p
                       in itertools.chain(self.loader_params.items(),
                                          user_params.items())}
        self.meta = self.loader.load_section('meta', params=meta_params)
        self.sns_topic_arn = self.environment.sns_topic_arn
        self.tags = {
            'humilis:environment': self.env_name,
            'humilis:layer': self.name,
            'humilis:stage': self.env_stage,
            'humilis:created': str(datetime.datetime.now())}
        for tagname, tagvalue in self.environment.tags.items():
            self.tags[tagname] = tagvalue

        for tagname, tagvalue in self.meta.get('tags', {}).items():
            self.tags[tagname] = tagvalue

        self.yaml_params = self.meta.get('parameters', {})
        for k, v in self.yaml_params.items():
            # Set 1 as default priority for all parameters
            v['priority'] = v.get('priority', 1)

        # User params override what is in the layer definition file
        self.user_params = user_params
        for pname, pvalue in user_params.items():
            if pname in self.yaml_params:
                self.yaml_params[pname]['value'] = pvalue
        self.__ec2 = None
        self.__s3 = None

    @property
    def termination_protection(self):
        """Is termination protection set for this layer?."""
        return self.meta.get('parameters', {}).get(
            'termination_protection', {}).get('value', False)

    @property
    def cf_name(self):
        """The name of the CF stack associated to this layer."""
        return get_cf_name(self.env_name, self.name, stage=self.env_stage)

    @property
    def loader_params(self):
        """Produces a dictionary of parameters to pass to a section loader."""
        # User parameters in the layer meta.yaml
        # Not that some param values may not have been populated when this
        # property is accessed since that may happen during the parsing of
        # some references in the parameter list: thus the if 'value' in v

        # For backwards compatibility, to be deprecated
        params = {k: v['value'] for k, v in self.params.items()
                  if 'value' in v}

        params["__vars"] = dict(params)

        # For backwards compatibility, to be deprecated
        params['_env'] = {'stage': self.env_stage, 'name': self.env_name,
                          'basedir': self.env_basedir}
        params['_os_env'] = os.environ
        params['_layer'] = {'name': self.name}
        params['env'] = os.environ

        # The new format:
        params['__env'] = os.environ
        params['__context'] = {
            'environment': {
                'name': self.env_name,
                'basedir': self.env_basedir
            },
            'stage': self.env_stage,
            'layer': {
                'name': self.name,
                'basedir': self.basedir
            },
            'aws': {
                'account_id': boto3.client('sts').get_caller_identity().get('Account')
            }
        }

        # For backwards compatibility
        params["context"] = params["__context"]
        return params

    @property
    def in_cf(self):
        """Returns true if the layer has been already deployed to CF."""
        return self.cf_name in {stk['StackName'] for stk in self.cf.stacks}

    @property
    def ec2(self):
        """Connection to AWS EC2 service."""
        if self.__ec2 is None:
            self.__ec2 = Ec2(config.boto_config)
        return self.__ec2

    @property
    def s3(self):
        """Connection to AWS S3."""
        if self.__s3 is None:
            self.__s3 = S3(config.boto_config)
        return self.__s3

    @property
    def ok(self):
        """Layer is fully deployed in CF and ready for use"""
        return self.cf.stack_ok(self.cf_name)

    @property
    def outputs(self):
        """Layer CF outputs."""
        ly = self.cf.stack_outputs.get(self.cf_name)
        if ly:
            ly = {o['OutputKey']: o['OutputValue'] for o in ly}
        return ly

    @property
    def resources(self):
        """Layer CF resources."""
        ly = self.cf.get_stack_resources(self.cf_name)
        if ly:
            ly = {o.logical_id: o.physical_resource_id for o in ly}
        return ly

    def compile(self):
        """Loads all files associated to a layer."""
        # Some templates may refer to params, so populate them first
        self.populate_params()

        # Load all files with layer contents
        for section in config.LAYER_SECTIONS:
            self.section[section] = self.loader.load_section(
                section, params=self.loader_params)

        # Package the layer as a CF template
        default_description = "{}-{} ({})".format(
            self.environment.name, self.name, self.environment.stage)
        description = self.params.get('description', {}).get('value') or \
                self.meta.get('description') or \
                default_description
        cf_template = {
            'AWSTemplateFormatVersion': str(config.CF_TEMPLATE_VERSION),
            'Description': description,
            'Mappings': self.section.get('mappings', {}),
            'Parameters': self.section.get('parameters', {}),
            'Resources': self.section.get('resources', {}),
            'Outputs': self.section.get('outputs', {})
        }
        if self.section.get('transform', {}).get('value', {}):
            cf_template['Transform'] = self.section['transform']['value']
        return cf_template

    def populate_params(self):
        """Populates parameters in a layer by resolving references."""
        if len(self.yaml_params) < 1:
            return
        for pname, param in sorted(self.yaml_params.items(),
                                   key=lambda t: t[1].get('priority', '1')):
            self.params[pname] = {}
            self.params[pname]['description'] = param.get('description', None)
            try:
                self.params[pname]['value'] = self._parse_param_value(
                    param['value'])
            except:
                self.logger.error("Error parsing layer '{}'".format(self.name))
                raise

    def print_params(self):
        """Prints the params used during layer creation."""
        if len(self.params) < 1:
            print("No parameters. Did you forget to run populate_params()?")
            return
        print("Parameters for layer {}:".format(self.cf_name))
        for pname, param in self.params.items():
            pval = param.get('value', None)
            if len(pval) > 30:
                pval = pval[0:30]
            print("{pname:<15}: {pval:>30}".format(pname=pname, pval=pval))

    def _parse_param_value(self, pval):
        """Parses layer parameter values."""
        if isinstance(pval, list):
            return [self._parse_param_value(_) for _ in pval]
        elif _is_reference(pval):
            return self._resolve_ref(
                list(pval.keys())[0][1:], list(pval.values())[0])
        elif _is_legacy_reference(pval):
            return self._resolve_ref(
                pval['ref']['parser'], pval['ref'].get('parameters', {}))
        elif isinstance(pval, dict):
            return {k: self._parse_param_value(v) for k, v in pval.items()}
        else:
            return pval

    def _resolve_ref(self, parsername, parameters):
        """Resolves references."""
        parser = config.reference_parsers.get(parsername)
        if not parser:
            msg = "Invalid reference parser '{}' in layer '{}'".format(
                parsername, self.cf_name)
            raise ReferenceError(ref, msg, logger=self.logger)
        result = parser(self, config.boto_config, **parameters)
        return result

    def delete(self):
        """Deletes a stack in CF."""
        msg = "Deleting stack {} from CF".format(self.cf_name)
        self.logger.info(msg)
        self.cf.delete_stack(self.cf_name)

    def create(self, update=False):
        """Deploys a layer as a CF stack."""
        msg = "Starting checks for layer {}".format(self.name)
        self.logger.info(msg)

        # CAPABILITY_IAM is needed only for layers that contain certain
        # resources, but we add it  always for simplicity.
        if not self.in_cf:
            self.logger.info("Creating layer '{}' (CF stack '{}')".format(
                self.name, self.cf_name))

            cf_template = self.compile()
            try:
                self.create_with_changeset(cf_template)
            except Exception:
                self.logger.error(
                    "Error deploying stack '{}'".format(self.cf_name))
                self.logger.error("Stack template: {}".format(
                    json.dumps(cf_template, indent=4)))
                raise
        elif update:
            cf_template = self.compile()
            try:
                self.create_with_changeset(cf_template, update)
            except NoUpdatesError:
                msg = "Nothing to update on stack '{}'".format(self.cf_name)
                self.logger.warning(msg)
            except Exception:
                self.logger.error(
                    "Error deploying stack '{}'".format(self.cf_name))
                self.logger.error("Stack template: {}".format(
                    json.dumps(cf_template, indent=4)))
                raise
        else:
            msg = "Layer '{}' already in CF: not creating".format(self.name)
            self.logger.info(msg)

        return self.outputs

    def _upload_cf_template(self, cf_template):
        """Upload CF template to S3."""
        bucket = config.boto_config.profile.get('bucket')
        key = "{}{}-{}.json".format(
            self.s3_prefix, round(time.time()), str(uuid.uuid4()))
        cf_template = json.dumps(cf_template).encode()
        S3().resource.Bucket(bucket).put_object(Key=key, Body=cf_template)
        return "https://s3-{}.amazonaws.com/{}/{}".format(
            config.boto_config.profile['aws_region'], bucket, key)

    def create_with_changeset(self, cf_template, update=False):
        """Use a changeset to create a stack."""
        changeset_type = "CREATE"
        if update:
            changeset_type = "UPDATE"
        changeset_name = self.cf_name + str(uuid4())
        template_url = self._upload_cf_template(cf_template)
        self.cf.client.create_change_set(
            StackName=self.cf_name,
            TemplateURL=template_url,
            Capabilities=["CAPABILITY_IAM", "CAPABILITY_NAMED_IAM"],
            NotificationARNs=self.sns_topic_arn,
            Tags=[{"Key": k, "Value": v} for k, v in self.tags.items()],
            ChangeSetName=changeset_name,
            ChangeSetType=changeset_type)
        self.wait_for_status_change()
        self.wait_changeset_creation(changeset_name)
        if update:
            changeset = self.cf.client.describe_change_set(
                ChangeSetName=changeset_name,
                StackName=self.cf_name)
            if not changeset["Changes"]:
                raise NoUpdatesError("Nothing to update")
        self.cf.client.execute_change_set(ChangeSetName=changeset_name,
                                          StackName=self.cf_name)
        self.wait_for_status_change()

    @staticmethod
    def _is_bad_status(status):
        """True if a stack status is not healthy."""
        return status is None \
                or status not in {"CREATE_COMPLETE", "UPDATE_COMPLETE",
                                  "REVIEW_IN_PROGRESS",
                                  "UPDATE_ROLLBACK_COMPLETE"}

    def _print_events(self, already_seen=None):
        """Prints the events reported by AWS."""
        if already_seen is None:
            already_seen = set()

        events = self.cf.get_stack_events(self.cf_name)
        new_events = [ev for ev in events if ev.id not in already_seen]
        cm = config.EVENT_STATUS_COLOR_MAP
        for event in new_events:
            self.logger.info(
                "{color}{status}\033[0m {restype} {logid} "
                "{reason}".format(
                    color=cm.get(event.resource_status, ''),
                    status=event.resource_status,
                    restype=event.resource_type,
                    logid=event.logical_resource_id,
                    reason=event.resource_status_reason or "",
                ))
            already_seen.add(event.id)

        return already_seen

    def wait_for_status_change(self):
        """Wait for the status deployment state to change."""
        status, seen_events = self.watch_events()
        if self._is_bad_status(status):
            # One retry, also to flush all events
            status, seen_events = self.watch_events(already_seen=seen_events)
            if self._is_bad_status(status):
                # One retry
                msg = "Unable to deploy layer '{}': status is {}".format(
                    self.name, status)
                raise CloudformationError(msg, logger=self.logger)
        return status

    def watch_events(self,
                     progress_status={'CREATE_IN_PROGRESS',
                                      'UPDATE_IN_PROGRESS',
                                      'UPDATE_COMPLETE_CLEANUP_IN_PROGRESS'},
                     already_seen=None):
        """Watches CF events during stack creation."""
        stack_status = self.cf.get_stack_status(self.cf_name)
        if already_seen is None:
            already_seen = set()
        while (stack_status is None) or (stack_status in progress_status):
            already_seen = self._print_events(already_seen)
            time.sleep(5)
            stack_status = self.cf.get_stack_status(self.cf_name)

        return stack_status, already_seen

    def wait_changeset_creation(self, changeset_name,
                                progress_status={"CREATE_PENDING",
                                                 "CREATE_IN_PROGRESS"}):
        """Wait for a changeset to be in the right status to be executed."""
        status = self.cf.client.describe_change_set(
            ChangeSetName=changeset_name, StackName=self.cf_name)["Status"]
        while (status is None) or (status in progress_status):
            status = self.cf.client.describe_change_set(
                ChangeSetName=changeset_name, StackName=self.cf_name)["Status"]
            time.sleep(5)
        if status != "CREATE_COMPLETE":
            msg = "Unable to deploy layer '{}': changeset status is {}".format(
                self.name, status)
            raise CloudformationError(msg, logger=self.logger)
        return status

    def __repr__(self):
        return str(self)

    def __str__(self):
        args = re.sub(r'\'(\w+)\'\s*:\s*', r'\1=', str(self.user_params))[1:-1]
        if len(args) > 0:
            basestr = "Layer({env}, '{name}', {args})"
        else:
            basestr = "Layer({env}, '{name}')"
        return basestr.format(
            env=self.__environment_repr, name=self.name, args=args)
